#!/usr/bin/env python3
"""
Evaluate completion function of QLever
"""

from typing import TextIO, Union, Dict, Optional, List, Any, Set, Tuple as T  # noqa: F401, E501 pylint: disable=unused-import
import sys
import random
import traceback
import os.path
import re

import argparse
from urllib.parse import urlparse
import yaml

from backend import qlever, blazegraph, dummy, Backend, BackendError  # noqa: E501 pylint: disable=unused-import
from query_model import parse_sparql
from helpers import Tee, Colors
from sparql_completer import SparqlCompleter, NamedQuery
from eval_data import QueryRecord, EvalData
import template_handling


def complete_all(queries_yaml: Dict[str, Any],  # noqa: E501 pylint: disable=too-many-locals
                 num_skipped: int,
                 output_basename: str,
                 comp_templates_yaml: Dict[str, Any],
                 fallback_comp_templates_yaml: Dict[str, Any],
                 output_path: str,
                 show_queries: bool,
                 backend: qlever.Backend,
                 fallback_backend: qlever.Backend,
                 clear_cache_after_query: bool,
                 agnostic_templates_yaml = None,
                 agnostic_fallback_timeout = None,
                 divArgs = None):
    """
    Recreates the given queries using auto completion
    """
    completion_query_log_path = os.path.join(output_path,
            output_basename + '.completion.executed.yml')
    eval_log_path = os.path.join(output_path,
            output_basename + '.completion.results.tsv')
    fail_list_path = os.path.join(output_path,
            output_basename + '.fail.txt')
    with open(completion_query_log_path,
              'wt', encoding='utf-8') as completion_query_log, \
            open(eval_log_path,
                 'wt', encoding='utf-8') as eval_log, \
            open(fail_list_path, 'wt', encoding='utf-8') as fail_log:

        is_mixed_mode = agnostic_templates_yaml is not None
        comp_templates = {}
        comp_agnostic_templates = {}
        fallback_comp_templates = {}
        replace_predicates = {}  # type: Dict[str, str]

        temps = [comp_templates, fallback_comp_templates, comp_agnostic_templates] if is_mixed_mode else [comp_templates, fallback_comp_templates]
        temps_yaml = [comp_templates_yaml, fallback_comp_templates_yaml, agnostic_templates_yaml] if is_mixed_mode else [comp_templates_yaml, fallback_comp_templates_yaml]
        print()
        print("Reading completion templates:")
        for yaml_templates, template_target in zip(temps_yaml, temps):
            for comp_template in yaml_templates['queries']:
                query_name = str(comp_template['query'])
                # we cannot parse the templates as sparql, since they contain conditions etc.
                # TODO<joka921>:
                # if we apply all valid replacements here for test we could do a sanity check again
                print('Query:', query_name)
                comp = comp_template['sparql']
                template_target[comp_template['query']] = comp
                if show_queries:
                    print('\n'+comp)
            # also add the warmup queries
            if 'warmup-queries' in yaml_templates:
                for comp_query_yaml in yaml_templates['warmup-queries']:
                    query_name = str(comp_query_yaml['query'])
                    template_target[query_name] =comp_query_yaml['sparql']

            # and also the global prefixes
            if 'global-prefixes' in yaml_templates:
                template_target['global-prefixes'] = yaml_templates['global-prefixes']

            """
            if 'replace-predicates-in-context' in comp_templates_yaml:
                replace_predicates = \
                    comp_templates_yaml['replace-predicates-in-context']
            """


        completer = SparqlCompleter(comp_templates,
                                    replace_predicates,
                                    completion_query_log,
                                    eval_log, show_queries, backend,
                                    fallback_backend, fallback_comp_templates, comp_agnostic_templates,
                                    agnostic_fallback_timeout,
                                    divArgs
                                    )

        completer.eval.num_skipped = num_skipped
        completer.eval.write_log_header()
        # first get all aliases to not let them interfere with anything
        alias_maps = dict()
        for query_yaml in queries_yaml['queries']:
            query_name = str(query_yaml['query'])
            query = query_yaml['sparql']

            alias_maps[query_name] = completer._create_alias_map(query.where)
        # clear cache to get rid of alias query artifacts
        backend.clear_server_cache()
        if divArgs.mode != "sensitive":
            fallback_backend.clear_server_cache()

        for query_yaml in queries_yaml['queries']:
            if clear_cache_after_query:
                backend.clear_server_cache()
                fallback_backend.clear_server_cache()
            query_name = str(query_yaml['query'])
            print()
            print(Colors.BOLD + 'Query:', query_name + Colors.ENDC)
            query = query_yaml['sparql']

            #print('Clearing backend cache (without pinned)')
            #backend.clear_server_cache()
            #fallback_backend.clear_server_cache()

            if not completer.complete(query, alias_maps[query_name]):
                completer.eval.num_failed += 1
                fail_log.write(query.name+'\n')
                continue
            print(Colors.OKGREEN, end="")
            print('Completed whole target!')
            print(Colors.ENDC, end="")

        print()
        completer.eval.print()


def completion_eval(backend: Backend, fallback_backend:Backend, queries: Dict[str, Any],
                    num_skipped: int,
                    output_basename: str,
                    output_path: str,
                    comp_templates_path: str,
                    fallback_templates_path: str,
                    show_queries: bool,
                    no_warmup: bool,
                    clear_cache_after_query: bool,
                    agnostic_templates_path = None,
                    agnostic_timeout = None,
                    divArgs = None
                    ):
    """
    Evaluate the completion functionality by completing all queries
    in the given query list (directly loaded from the YAML format)
    """

    if not comp_templates_path:
        comp_templates_path = queries['kb'] + \
            '.completion-templates.yml'

    log_path = os.path.join(output_path, output_basename + '.log')

    with Tee(log_path, 'wt', encoding='utf-8'):
        try:
            comp_templates = handle_warmup(comp_templates_path, queries, no_warmup, backend, clear_cache_after_query, show_queries)
            if divArgs.mode != "sensitive":
                fallback_comp_templates = handle_warmup(fallback_templates_path, queries, no_warmup, fallback_backend, clear_cache_after_query, show_queries)
            else:
                fallback_comp_templates = dict()
                fallback_comp_templates["queries"] = []
            agnostic_templates = None
            if agnostic_templates_path is not None and divArgs.mode != "sensitive" :
                agnostic_templates = handle_warmup(agnostic_templates_path, queries, no_warmup, fallback_backend, clear_cache_after_query,
                                               show_queries)

            complete_all(queries,
                         num_skipped,
                         output_basename,
                         comp_templates,
                         fallback_comp_templates,
                         output_path,
                         show_queries,
                         backend,
                         fallback_backend,
                         clear_cache_after_query,
                         agnostic_templates,
                         agnostic_timeout,
                         divArgs
                         )
        except Exception as e:
            print(str(e))
            print(traceback.format_exc())



def query_execution_eval(backend: Backend, queries: Dict[str, Any],
                         num_skipped: int,
                         output_basename: str,
                         output_path: str):
    query_log_path = os.path.join(output_path,
                                  output_basename + '.queries.executed.yml')
    eval_log_path = os.path.join(output_path,
                                 output_basename + '.queries.results.tsv')
    fail_list_path = os.path.join(output_path,
                                  output_basename + '.queries.fail.txt')
    with open(query_log_path,
              'wt', encoding='utf-8') as query_log, \
            open(eval_log_path,
                 'wt', encoding='utf-8') as eval_log, \
            open(fail_list_path,
                 'wt', encoding='utf-8') as fail_log:
        evaluator = EvalData(
            QueryRecord,
            eval_log,
            query_log)
        evaluator.num_skipped = num_skipped
        evaluator.write_log_header()
        for query_yaml in queries['queries']:
            query = query_yaml['sparql']
            evaluator.current_record.query_name = query_yaml['query']
            print('Query:', evaluator.current_record.query_name)
            evaluator.current_record.query = query

            try:
                sparql = query if type(query) == str else query.to_sparql()
                result, rti = backend.query(sparql)
                if result is None:
                    fail_log.write(evaluator.current_record.query_name+'\n')
                    evaluator.num_failed += 1
            except BackendError:
                fail_log.write(evaluator.current_record.query_name+'\n')
                evaluator.current_record.query = query

            evaluator.current_record.runtime_info = rti
            evaluator.current_record.query_client_time = \
                backend.last_query_client_time
            evaluator.current_record.query_time = \
                backend.last_query_client_time
            if backend.last_query_compute_time:
                # Override query_time with server measured time if available
                evaluator.current_record.query_time = \
                    backend.last_query_compute_time
            evaluator.finish_log_record()
            sys.stdout.flush()

        print(Colors.OKGREEN)
        print('Completed all queries!')
        print(Colors.ENDC)
        evaluator.print()

def handle_warmup(comp_templates_path, queries, no_warmup, backend, clear_cache_after_query, show_queries):
    with open(comp_templates_path, 'rt', encoding='utf-8') \
            as comp_templates_file:
        comp_templates = yaml.safe_load(comp_templates_file)

        if queries['kb'] != comp_templates['kb']:
            print('Query KB {} does not match completion KB {}'.format(
                queries['kb'], comp_templates['kb']))

        global_prefixes = ""
        if 'global-prefixes' in comp_templates:
            global_prefixes = comp_templates["global-prefixes"]

        if 'warmup-queries' in comp_templates and not no_warmup:
            replacements = dict()
            for comp_query_yaml in comp_templates['warmup-queries']:
                query_name = str(comp_query_yaml['query'])
                replacements[query_name.upper()] =comp_query_yaml['sparql']
            print()
            print('Executing warmup queries:')
            for comp_query_yaml in comp_templates['warmup-queries']:
                if clear_cache_after_query:
                    backend.clear_server_cache()

                query_name = str(comp_query_yaml['query'])
                print('Query:', query_name, end="")
                if show_queries:
                    print()
                else:
                    print(" ... ", end="")
                    sys.stdout.flush()
                sparql = template_handling.substitute(comp_query_yaml['sparql'], replacements, True)
                sparql = global_prefixes + '\n' + sparql
                query = NamedQuery(parse_sparql(sparql, True))
                query.name = query_name
                if show_queries:
                    print('-'*80)
                    print(query.to_sparql())
                    print('-'*80)
                # let the warmup queries never timeout
                _, _, time = backend.query(query.to_sparql(), pin_result=True, timeout=200)
                if (time.client_time is not None or time.query_time is not None):
                    print('%dms' % (int(time.query_time * 1000)
                                    if time.query_time
                                    else time.client_time))
                else:
                    print('no query or client time returned from backend')
    return comp_templates

def main(argv: List[str]):
    """
    Main method handling the creation of parsers, backend connections etc
    """
    random.seed(42)
    # Keep dict order in YAML instead of sorting alphabetically.
    # This keeps the runtime information readable

    parser = argparse.ArgumentParser(
        description='Evaluate QLever\'s auto completion')
    parser.add_argument('evaluation', type=str.lower,
                        default='completion',
                        choices=['completion', 'queries'],
                        help='Whether to evaluate completion or the queries')

    parser.add_argument('--qlever-backend', type=str,
                        help='URL of a QLever instance (always needed for the warmed-up queries)',
                        default='http://vulcano.informatik.privat:9001')
    parser.add_argument('--blazegraph-backend', type=str,
                        help='URL of a Blazegraph instance (unneeded if backend-type is qlever or virtuoso)',
                        default='http://vulcano.informatik.privat:9001')
    parser.add_argument('--virtuoso-backend', type=str,
                        help='URL of a virtuoso instance (unneeded if backend-type is qlever or blazegraph)',
                        default='http://vulcano.informatik.privat:9001')
    parser.add_argument('--exclude', type=str,
                        help='Path to a line separated list of query names'+
                        ' that should be excluded', default=None)
    parser.add_argument('--qlever-completion-templates', type=str, default='')
    parser.add_argument('--blazegraph-completion-templates', type=str, default='')
    parser.add_argument('--virtuoso-completion-templates', type=str, default='')
    parser.add_argument('--agnostic-qlever-templates', type=str, default='')
    parser.add_argument('--unranked-qlever-templates', type=str, default='')
    parser.add_argument('--mode', default='sensitive', choices=['sensitive','agnostic', 'unranked', 'mixed'])
    parser.add_argument('--mixed-timeout',type=float, default=5)
    parser.add_argument('--reverse-fixed-object-triples', type=str, choices=["True", "False"], default=False)
    parser.add_argument('--output-path', dest='output_path', type=str, default='output')
    parser.add_argument('--output-basename', dest='output_basename', type=str, default='')
    parser.add_argument('--force-parsing', dest='force_parsing',
                        action='store_true',
                        help='Force parsing of queries even if it\'s' +
                        'not necessary (e.g. when just executing)')
    parser.set_defaults(force_parsing=False)


    parser.add_argument('--initial-clear-cache', dest='clear_cache_initial',
                        action='store_true',
                        help='Clear cache before first query')
    parser.set_defaults(clear_cache_initial=True)

    parser.add_argument('--clear-cache-after-query',
                         dest='clear_cache_after_query',
                         help='clear the cache (not the pinned cache if' +
                         'supported by backend) after each query was completed)',
                         action='store_true', default=False)
    parser.add_argument('--no-warmup', dest='no_warmup',
                        action='store_true',
                        help='Do not execute warmup queries')
    parser.set_defaults(no_warmup=False)
    parser.add_argument('--clear-cache', dest='clear_cache',
                        action='store_true', help='Clear cache before each completion query')
    parser.add_argument('--backend-type', default='qlever',
                        dest='backend_type',
                        choices=['qlever', 'blazegraph', 'virtuoso', 'dummy'], type=str.lower,
                        help='The backend to use')
    parser.add_argument('--timeout-seconds', dest='timeout_seconds', type=int, default=120,
                        help='Timeout completion queries in the primary backend after this many seconds')
    parser.add_argument('--show-queries', dest='show_queries',
                        action='store_true',
                        help='Show SPARQL queries in console log',
                        default=False)
    parser.add_argument('--test-mode', dest='test_mode', type=int, default=-1,
                        help='Do not removed pinned queries and stop after'
                        'specified number of target queries')
    parser.add_argument('--single-query', dest='single_query',
                        type=str, default='',
                        help='Evaluate only the given query')
    parser.set_defaults(clear_cache=False)
    parser.add_argument('queries', type=str)
    parser.add_argument('--exclude-labels', type=str, default="False", choices=["True", "False"], help="Don't autocomplete triples where the predicate sth like rdfs:label")
    parser.add_argument('--typing-steps', type=int, default=[0, 3, 7], nargs="+", help="For each token issue a query for each prefix length in this list. If omitted, use the legacy mode (2 chars at least, then one char at a time until success")
    parser.add_argument('--query-variables', type=str, default="False", choices=["True", "False"], help="Also issue a completion query for variables and wait until it returns")
    parser.add_argument('--certify-fallback', type=str, default="False", choices=["True", "False"], help="when using the agnostic falllback, verify for the first results, whether they lead to a non-empty result")
    parser.add_argument('--superset-predicates', type=str, default="False", choices=["True", "False"], help="When completing an object and the predicate is an expensive property path, replace it by a simpler superset")

    args = parser.parse_args(argv[1:])

    if args.blazegraph_backend.endswith('/'):
        args.blazegraph_backend = args.blazegraph_backend[:-1]
    # A class to store several arguments that have to be passed down to the actual
    # completion functions

    # unranked works exactly the same as agnostic, just with different templates
    if args.mode == "unranked":
        args.agnostic_qlever_templates = args.unranked_qlever_templates
        args.mode = "agnostic"
    class ArgStore:
        pass
    divArgs = ArgStore()
    divArgs.exclude_labels=args.exclude_labels=="True"
    divArgs.forced_typing_steps=args.typing_steps
    divArgs.query_variables=args.query_variables=="True"
    divArgs.reverse_fixed_object_triples=args.reverse_fixed_object_triples=="True"
    divArgs.certify_fallback = args.certify_fallback == "True"
    divArgs.mode = args.mode

    divArgs.superset_predicates = args.superset_predicates == "True"

    qlever_backend_url = urlparse(args.qlever_backend)
    qlever_backend_host, qlever_backend_port = qlever_backend_url.netloc.split(':')
    qlever_backend_path = qlever_backend_url.path if qlever_backend_url.path else '/'

    qlever_ssl = args.qlever_backend.strip().startswith("https:") or qlever_backend_port == 443

    if not qlever_backend_port:
        qlever_backend_port = 80

    qlever_backend = qlever.Qlever(qlever_backend_host,
                            qlever_backend_port,
                            qlever_backend_path,
                            default_timeout_seconds=args.timeout_seconds,
                            clear_cache=args.clear_cache, ssl=qlever_ssl)

    backend = None  # type: Optional[Backend]
    templates = args.qlever_completion_templates
    if args.backend_type == 'qlever':
        backend = qlever_backend
    elif args.backend_type in ['blazegraph', 'virtuoso']:
        relevant_backend = args.blazegraph_backend if args.backend_type == 'blazegraph' else args.virtuoso_backend
        backend_url = urlparse(relevant_backend)
        backend_host, backend_port = backend_url.netloc.split(':')
        backend_path = backend_url.path if backend_url.path else '/'
        if not backend_port:
            backend_port = 80
        ssl = relevant_backend.strip().startswith("https:") or backend_port == 443
        backend = blazegraph.Blazegraph(backend_host,
                                        backend_port,
                                        backend_path,
                                        backend_timeout_seconds=args.timeout_seconds,
                                        clear_cache=args.clear_cache, ssl=ssl)
        if args.backend_type == 'virtuoso':
            backend.is_blazegraph = False
        templates = args.blazegraph_completion_templates if args.backend_type == 'blazegraph' else args.virtuoso_completion_templates
    else:
        assert (args.backend_type == 'dummy')
        backend = dummy.DummyBackend(0, 0, 0, clear_cache=args.clear_cache)
        qlever_backend = backend

    # Read set of excluded queries.
    exclude_set = set()
    if args.exclude:
        with open(args.exclude, 'rt', encoding='utf-8') as exc:
            for line in exc:
                exclude_set.add(line.strip())

    # Read queries and skip some (see comment below).
    if (args.output_basename == ""):
        assert (args.queries.endswith("queries.yml"))
        args.output_basename = re.sub(".queries.yml$", "", os.path.basename(args.queries))
    parse_fail_list = []
    with open(args.queries, 'rt', encoding='utf-8') as queries_file:
        queries_yaml = yaml.safe_load(queries_file)
        filtered_queries = []  # type: List[Union[str, NamedQuery]]
        num_skipped = 0
        num_queries = 0
        for query_yaml in queries_yaml['queries']:
            num_queries += 1
            query_name = query_yaml['query']

            # Exclude query if:
            # 1. Query name is in exclude set
            # 2. Single query mode and name != single query name
            # 3. Test mode > 0 and query_count > test mode count
            if (query_name in exclude_set) or \
               (args.single_query != "" and query_name != args.single_query) or \
               (args.test_mode > 0 and num_queries > args.test_mode):
                num_skipped += 1
                continue

            orig_sparql = query_yaml['sparql']
            # We always need to parse for the completion
            if args.force_parsing or args.evaluation == 'completion':
                try:
                    # currently disable property paths
                    query = NamedQuery(parse_sparql(orig_sparql, True))
                    query.name = query_name
                    query_yaml['sparql'] = query
                    filtered_queries.append(query_yaml)
                except (ValueError, NotImplementedError) as ex:
                    if str(ex):
                        print(str(ex))
                    print(Colors.FAIL +
                          'Failed to parse query: ' + query_name + '\n' +
                          Colors.ENDC)
                    num_skipped += 1
                    parse_fail_list.append(query_name+'\n')
                    continue
            else:
                filtered_queries.append(query_yaml)
        queries_yaml['queries'] = filtered_queries
        print()
        print("Queries skipped: %d of %d" % (num_skipped, num_queries))

        # Dump the input queries with excluded removed
        # and possibly simplified (if args.force_parsing)
        queries_path = os.path.join(args.output_path,
                                    args.output_basename +
                                    '.input-queries.yml')
        with open(queries_path, 'wt', encoding='utf-8') as queries_stream:
            yaml.dump(queries_yaml, queries_stream)

        if parse_fail_list:
            parse_fail_path = os.path.join(args.output_path,
                                           args.output_basename +
                                           '.parse-fail.txt')
            with open(parse_fail_path, 'wt', encoding='utf-8') \
                    as parse_fail_file:
                parse_fail_file.writelines(parse_fail_list)

        # Initial clear cache (also removed pinned results, unles in test mode)
        #
        # NOTE: This came earlier in the code, but I didn't see a reason to do
        # this before the parsing, especially since it can incur an unnecessary
        # wait if something goes wrong in the previous steps.
        if args.clear_cache_initial:
            print()
            if args.test_mode == -1:
                print('Clearing backend cache (including pinned results)')
                backend.clear_server_cache(True)
                if (divArgs.mode != "sensitive"):
                    qlever_backend.clear_server_cache(True)
            else:
                print('Clearing backend cache (but leaving pinned results)')
                backend.clear_server_cache(False)
                if (divArgs.mode != "sensitive"):
                    qlever_backend.clear_server_cache(False)

        # Below we need to forward the number of queries skipped at this point
        # as these queries are now gone.  On the other hand the completion or
        # query execution steps may still skip some more queries.


        if args.evaluation == 'completion':
            timeout = None # None signals sensitive mode
            if args.mode=="mixed":
                timeout = args.mixed_timeout
            if args.mode=="agnostic":
                timeout = 0  # immediately go to the fallback backend
            completion_eval(backend, qlever_backend, queries_yaml,
                            num_skipped,
                            args.output_basename,
                            args.output_path,
                            templates,
                            args.qlever_completion_templates,
                            args.show_queries,
                            args.no_warmup,
                            args.clear_cache_after_query,
                            args.agnostic_qlever_templates,
                            timeout, divArgs)
        else:
            query_execution_eval(backend, queries_yaml,
                                 num_skipped,
                                 args.output_basename,
                                 args.output_path)


if __name__ == '__main__':
    main(sys.argv)
