#!/usr/bin/env python3
"""
Module for some nice helper functions for evaluation
"""
import sys


class Tee:
    """
    Funny little helper class that allows us to redirect all output to a log
    file and also to stdout
    """
    def __init__(self, name, *args, **kwargs):
        self.file = open(name, *args, **kwargs)
        self.stderr = sys.stderr


        self.stdout = sys.stdout
        sys.stdout = self
        sys.stderr = self

    def close(self):
        """
        Acts like stdout.close() but also closes the log file
        """
        self.flush()
        if self.stdout is not None:
            sys.stdout = self.stdout
        if self.stderr is not None:
            sys.stderr = self.stderr
            self.stdout = None
        if self.file is not None:
            self.file.close()
            self.file = None

    def write(self, data):
        """
        Acts like stdout.write() but also writes to log
        """
        self.file.write(data)
        self.stdout.write(data)

    def flush(self):
        """
        Acts like stdout.flush() but also flushes the log file
        """
        if self.file is not None:
            self.file.flush()
        if self.stdout is not None:
            self.stdout.flush()
        if self.stderr is not None:
            self.stderr.flush()

    def __enter__(self):
        return self

    def __exit__(self, ex_type, ex_value, traceback):
        """
        Close on del
        """
        self.close()


class Colors:  # pylint: disable=too-few-public-methods
    """
    ANSI color sequence constants for nice printing
    """
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKGREEN = '\033[35m'
    HIGHLIGHT = '\033[34m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'
