"""
The query model module provides a simple data model allowing easier
manipulation of SPARQL queries
"""
from antlr4 import CommonTokenStream, InputStream
from antlr4.error.ErrorListener import ErrorListener

from query_model.SparqlLexer import SparqlLexer
from query_model.SparqlParser import SparqlParser
from query_model.query_builder import QueryBuilder
from query_model.query_model import SelectQuery


class RaiseErrorListener( ErrorListener ):
    """
    Turn ANTLR parser errors into ValueError exceptions
    """

    def syntaxError(self, recognizer, offendingSymbol, line, column, msg, e):
        print(str(line) + ":" + str(column) + ": syntax ERROR, " + str(msg))
        raise ValueError

    def reportAmbiguity(self, recognizer, dfa, startIndex, stopIndex, exact, ambigAlts, configs):
        print("Ambiguity ERROR, " + str(configs))
        raise ValueError

    def reportAttemptingFullContext(self, recognizer, dfa, startIndex, stopIndex, conflictingAlts, configs):
        print("Attempting full context ERROR, " + str(configs))
        raise ValueError

    def reportContextSensitivity(self, recognizer, dfa, startIndex, stopIndex, prediction, configs):
        print("Context ERROR, " + str(configs))
        raise ValueError


def parse_sparql(sparql: str, allowPropertyPaths: bool =False ) -> SelectQuery:
    """
    Parses SPARQL in a string to the query model using ANTLR
    """
    try:
        error_listener = RaiseErrorListener()
        inp = InputStream(sparql)
        lexer = SparqlLexer(inp)
        lexer.addErrorListener(error_listener)
        stream = CommonTokenStream(lexer)
        parser = SparqlParser(stream)
        parser.addErrorListener(error_listener)
        tree = parser.query()
        query_builder = QueryBuilder(allowPropertyPaths)
        query = query_builder.visitQuery(tree)
        return query
    except:
        print(sparql)
        raise

