"""
SPARQL AST visitor for creating an abstract SPARQL query model
"""
import re
from typing import Optional, List

from antlr4 import ParseTreeVisitor

import query_model
from query_model.SparqlParser import SparqlParser
from query_model.query_model import SelectQuery, Prologue, \
    Prefix, GroupClause, UnionPattern, \
    GroupGraphPattern, OptionalPattern, ValuesPattern, PropertyPath, BindPattern, Filter, Selector, MinusGraphPattern


class QueryBuilder(ParseTreeVisitor):

    def __init__(self, allowPropertyPaths=False):
        """
        if allowPropertyPaths, then property paths will be trivially parsed as their
        string representation. Else, all non-trivial (single-predicate) paths will throw an exception
        This is needed, since we currently have no datatypes representing propertyPaths
        :param allowPropertyPaths:
        """
        super().__init__()
        self.allowPropertyPaths = allowPropertyPaths
        self.prologue = None  # type: Optional[Prologue]

    # Visit a parse tree produced by SparqlParser#query.
    def visitQuery(self, ctx: SparqlParser.QueryContext) \
            -> SelectQuery:
        prologue = self.visitPrologue(ctx.prologue())
        if ctx.selectQuery():
            query = self.visitSelectQuery(ctx.selectQuery(), prologue)
        else:
            raise NotImplementedError("This parser only supports selectQueries")
        if ctx.valuesClause() and ctx.valuesClause().getText():
            print("values clause is: ")
            print(ctx.valuesClause().getText())
            raise NotImplementedError("No values clause after query is supported by this parser")
        return query

    # Visit a parse tree produced by SparqlParser#prologue.
    def visitPrologue(self, ctx: SparqlParser.PrologueContext):
        base = None
        if ctx.baseDecl():
            base = self.visitBaseDecl(ctx.baseDecl())

        prefixes = [] # type: List[Prefix]
        if ctx.prefixDecl():
            prefix_decls = ctx.prefixDecl()
            for prefix_decl in prefix_decls:
                prefixes.append(self.visitPrefixDecl(prefix_decl))

        return Prologue(base, prefixes)


    # Visit a parse tree produced by SparqlParser#baseDecl.
    def visitBaseDecl(self, ctx: SparqlParser.BaseDeclContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by SparqlParser#prefixDecl.
    def visitPrefixDecl(self, ctx: SparqlParser.PrefixDeclContext):
        return Prefix(ctx.PNAME_NS().getSymbol().text,
                      ctx.IRI_REF().getSymbol().text)

    def visitSelectClause(self, ctx: SparqlParser.SelectClauseContext):
        selectors = []  # type: List[Selector]
        distinct = bool(ctx.DISTINCT())
        if ctx.REDUCED():
            raise NotImplementedError("REDUCED is not supported by this parser")


        for child in ctx.children:
            if isinstance(child, query_model.SparqlParser.AliasContext):
                selectors.append(Selector(
                    '('+child.expression().getText()+' AS '+child.var().getText()+')',
                child.var().getText()))
            elif isinstance(child, query_model.SparqlParser.VarContext):
                selectors.append(Selector(child.getText(), child.getText()))
        return (distinct,  selectors)

    # Visit a parse tree produced by SparqlParser#selectQuery.
    def visitSelectQuery(self, ctx: SparqlParser.SelectQueryContext, prologue: Prologue) \
            -> SelectQuery:
        distinct, selectors = self.visitSelectClause(ctx.selectClause())
        if ctx.datasetClause():
            raise NotImplementedError("Dataset clauses in select queries are unsupported by this parser")
        where_clause = self.visitWhereClause(ctx.whereClause(), prologue)
        sol_mod = ctx.solutionModifier()
        order, limit, group = self.visitSolutionModifier(sol_mod)
        return SelectQuery(prologue, distinct, selectors, where_clause,
                           group, order, limit)

    # Visit a parse tree produced by SparqlParser#subSelect.
    def visitSubSelect(self, ctx:SparqlParser.SubSelectContext, prologue: Prologue):
        distinct, selectors = self.visitSelectClause(ctx.selectClause())
        where_clause = self.visitWhereClause(ctx.whereClause(), prologue)
        sol_mod = ctx.solutionModifier()
        order, limit, group = self.visitSolutionModifier(sol_mod)
        if ctx.valuesClause().getText() :
            raise NotImplementedError ("Values clauses after a subquery are not supported by this parser (subselect rule)  ")

        return SelectQuery(None, distinct, selectors, where_clause,
                           group, order, limit)

    def visitGroupClause(self, ctx: SparqlParser.GroupClauseContext):
        selectors = [cond.getText() for cond in ctx.groupCondition()]
        return GroupClause(selectors, None)

    # Visit a parse tree produced by SparqlParser#whereClause.
    def visitWhereClause(self, ctx: SparqlParser.WhereClauseContext, prologue: Prologue):
        graph_patterns = self.visitGroupGraphPattern(ctx.groupGraphPattern(), prologue)
        return graph_patterns

    # Visit a parse tree produced by SparqlParser#groupOrUnionGraphPattern.
    def visitGroupOrUnionGraphPattern(self, ctx: SparqlParser.GroupOrUnionGraphPatternContext, prologue: Prologue):
        groupGraphPatterns = ctx.groupGraphPattern()
        if not groupGraphPatterns:
            return None

        left = self.visitGroupGraphPattern(groupGraphPatterns[0], prologue)
        for groupGraphPattern in groupGraphPatterns[1:]:
            group_graph_pattern_model = self.visitGroupGraphPattern(groupGraphPattern, prologue)
            left = UnionPattern(left, group_graph_pattern_model)

        return left

    # Visit a parse tree produced by SparqlParser#groupGraphPattern.
    def visitGroupGraphPattern(self, ctx:SparqlParser.GroupGraphPatternContext, prologue: Prologue):
        if ctx.subSelect():
            return self.visitSubSelect(ctx.subSelect(), prologue)
        elif ctx.groupGraphPatternSub():
            return self.visitGroupGraphPatternSub(ctx.groupGraphPatternSub(), prologue)

    # Visit a parse tree produced by SparqlParser#groupGraphPattern.
    def visitGroupGraphPatternSub(self, ctx: SparqlParser.GroupGraphPatternContext, prologue: Prologue):
        patterns = GroupGraphPattern()
        for child in ctx.children:
            if isinstance(child, query_model.SparqlParser.GraphPatternNotTriplesContext):
                patterns.append(self.visitGraphPatternNotTriples(child, prologue))
            elif isinstance(child, query_model.SparqlParser.TriplesBlockContext):
                child_patterns = self.visitTriplesBlock(child)
                patterns.extend(child_patterns)
                # else case are fragments like the dot.
        return patterns

    def visitGraphPatternNotTriples(self, ctx: SparqlParser.GraphPatternNotTriplesContext, prologue: Prologue):
        for child in ctx.children:
            if isinstance(child, query_model.SparqlParser.GroupOrUnionGraphPatternContext):
                return self.visitGroupOrUnionGraphPattern(child, prologue)
            if isinstance(child, query_model.SparqlParser.OptionalGraphPatternContext):
                return self.visitOptionalGraphPattern(child, prologue)
            if isinstance(child, query_model.SparqlParser.FilterRContext):
                return self.visitFilter(child)
            if isinstance(child, query_model.SparqlParser.ServiceGraphPatternContext):
                return self.visitServiceGraphPattern(child)
            if isinstance(child, query_model.SparqlParser.InlineDataContext):
                return self.visitDataBlock(child.dataBlock())
            if isinstance(child, query_model.SparqlParser.BindContext):
                return BindPattern(child.expression().getText(), child.var().getText())
            if isinstance(child, query_model.SparqlParser.MinusGraphPatternContext):
                return MinusGraphPattern(self.visitGroupGraphPattern(child.groupGraphPattern(), prologue))


            raise NotImplementedError("Unsupported graphPatternType (GRAPH etc...")

        return None

    def visitDataBlock(self, ctx:SparqlParser.DataBlockContext):
        if ctx.inlineDataOneVar() :
            return self.visitInlineDataOneVar(ctx.inlineDataOneVar())
        else:
            return self.visitInlineDataFull(ctx.inlineDataFull())

    def visitInlineDataOneVar(self, ctx: SparqlParser.InlineDataOneVarContext):
        variable =[ctx.var().getText()]
        values = [[c.getText()] for c in ctx.dataBlockValue()]
        return ValuesPattern(variable, values)

    def visitInlineDataFull(self, ctx: SparqlParser.InlineDataFullContext):
        if ctx.NIL():
            raise ValueError("NIL values () in VALUE clauses are unsupported by this parser")
        variable =[c.getText() for c in ctx.var()]
        values = [[c.getText() for c in block.dataBlockValue()] for block in ctx.dataBlockSingle()]
        if not variable :
            raise ValueError("VALUES clauses that contain no variables are unsupported by this parser")
        return ValuesPattern(variable, values)



    # Visit a parse tree produced by SparqlParser#triplesBlock.
    def visitTriplesBlock(self, ctx: SparqlParser.TriplesBlockContext):
        triples = GroupGraphPattern()
        triples.extend(self.visitTriplesSameSubjectPath(ctx.triplesSameSubjectPath()))
        triples_block = ctx.triplesBlock()
        if triples_block:
            triples.extend(self.visitTriplesBlock(triples_block))
        return triples

    # Visit a parse tree produced by SparqlParser#filter.
    def visitOptionalGraphPattern(self, ctx: SparqlParser.OptionalGraphPatternContext, prologue: Prologue) -> OptionalPattern:
        groupGraphPattern = self.visitGroupGraphPattern(ctx.groupGraphPattern(), prologue)

        return OptionalPattern(groupGraphPattern)

    # Visit a parse tree produced by SparqlParser#filter.
    def visitFilter(self, ctx:SparqlParser.FilterRContext) -> Filter:
        constr = ctx.constraint().getText()
        vars = re.findall(r"\?[a-zA-Z_0-9]+", constr)
        return Filter('FILTER '+ctx.constraint().getText(), vars)

    # Visit a parse tree produced by SparqlParser#serviceGraphPattern.
    def visitServiceGraphPattern(self, ctx:SparqlParser.ServiceGraphPatternContext) -> str:
        return 'SERVICE wikibase:label { bd:serviceParam wikibase:language "en" . }'


    # Visit a parse tree produced by SparqlParser#triplesSameSubjectPath.
    def visitTriplesSameSubjectPath(self, ctx: SparqlParser.TriplesSameSubjectPathContext):
        subject = ctx.varOrTerm().getText()
        property_list = self.visitPropertyListPathNotEmpty(ctx.propertyListPathNotEmpty())
        triples = [[subject]+prop for prop in property_list]
        return triples


    # Visit a parse tree produced by SparqlParser#propertyListPathNotEmpty.
    def visitPropertyListPathNotEmpty(self, ctx:SparqlParser.PropertyListPathNotEmptyContext):
        props = []
        verb_nodes = iter(ctx.verbPathOrSimple())
        verb = self.visitVerbPathOrSimple(next(verb_nodes))
        objects_ctxs = self.visitObjectListPath(ctx.objectListPath())
        objects = [obj for obj in objects_ctxs]
        props.extend([[verb, obj] for obj in objects])
        for objects_node in ctx.objectList():
            verb = self.visitVerbPathOrSimple(next(verb_nodes))
            objects_ctx = self.visitObjectList(objects_node)
            objects = [obj for obj in objects_ctx]
            props.extend([[verb, obj] for obj in objects])
        return props

    def visitVerbPathOrSimple(self, ctx:SparqlParser.VerbPathOrSimpleContext):
        if ctx.verbSimple():
            return ctx.verbSimple().getText()
        else:
            return self.visitPathAlternative(ctx.verbPath().path().pathAlternative())

    def visitPathAlternative(self, ctx: SparqlParser.PathAlternativeContext, allow_multiple=True) :
        seq_nodes = ctx.pathSequence()
        alts = [self.visitPathSequence(n, allow_multiple) for n in seq_nodes]
        return PropertyPath(alts)

    def visitPathSequence(self, ctx: SparqlParser.PathSequenceContext, allow_multiple: bool) :
        nodes = ctx.pathEltOrInverse()
        els = [self.visitPathEltOrInverse(c, allow_multiple) for c in nodes]
        if len(els) > 1 and (not self.allowPropertyPaths or not allow_multiple)  :
            raise ValueError("Property paths are not allowed in autocompletion yet")
        return PropertyPath.PathSequence(els)

    def visitPathEltOrInverse(self, ctx: SparqlParser.PathEltOrInverseContext, allow_multiple: bool):
        isInverse = ctx.getText() and ctx.getText()[0] == '^'
        pathMod = ctx.pathElt().pathMod().getText() if ctx.pathElt().pathMod() else None
        primary = self.visitPathPrimary(ctx.pathElt().pathPrimary(), allow_multiple)

        if (isInverse or pathMod) and not self.allowPropertyPaths:
            raise ValueError("path modifiers and inverse properties currently not allowed for autocompletion")
        if pathMod and not allow_multiple:
            raise ValueError("no recursive inverse paths are currently allowed")
        if isinstance(primary, PropertyPath.PathElt):
            primary.modifier = pathMod
            return primary
        return PropertyPath.PathElt(primary, isInverse, pathMod)

    def visitPathPrimary(self, ctx: SparqlParser.PathPrimaryContext, allow_multiple:bool):
        if ctx.path():
            x = ctx.path()
            y = x.pathAlternative()

            return self.visitPathAlternative(y, False).alternatives[0].els[0]
        if not ctx.iri() :
            raise NotImplementedError("Only single iris are currently supported for autocompletion")
        return ctx.iri().getText()












        





    # Visit a parse tree produced by SparqlParser#propertyListNotEmpty.
    def visitPropertyListNotEmpty(self, ctx:SparqlParser.PropertyListNotEmptyContext):
        props = []
        verb_nodes = ctx.verb()
        object_lists = ctx.objectList()
        for verb_node, object_list_node in zip(verb_nodes, object_lists):
            verb = verb_node.getText()
            object_list = self.visitObjectList(object_list_node)
            objects = [obj for obj in object_list]
            props.extend([[verb, obj] for obj in objects])
        return props

    # Visit a parse tree produced by SparqlParser#objectList.
    def visitObjectList(self, ctx: SparqlParser.ObjectListContext):
        return [obj.getText() for obj in ctx.objectR()]

    # Visit a parse tree produced by SparqlParser#objectListPath.
    def visitObjectListPath(self, ctx: SparqlParser.ObjectListPathContext):
        return [obj.getText() for obj in ctx.objectPath()]

    # Visit a parse tree produced by SparqlParser#orderClause
    def visitOrderClause(self, ctx: SparqlParser.OrderClauseContext):
        return [cond.getText() for cond in ctx.orderCondition()]

    # Visit a parse tree produced by SparqlParser#solutionModifier
    def visitSolutionModifier(self, ctx: SparqlParser.SolutionModifierContext):

        order = None
        if ctx.orderClause():
            order = self.visitOrderClause(ctx.orderClause())
        limit = None
        if ctx.limitOffsetClauses():
            limit = self.visitLimitOffsetClauses(ctx.limitOffsetClauses())

        having = None
        if ctx.havingClause():
            having = ctx.havingClause().getText()

        #TODO: what does a having clause do if there is no groupBY?
        # we currently ignore it here
        group = None
        if ctx.groupClause():
            group = self.visitGroupClause(ctx.groupClause())
            group.having = having
        return order, limit, group




    # Visit a parse tree produced by SparqlParser#limitOffsetClauses
    def visitLimitOffsetClauses(self, ctx: SparqlParser.LimitOffsetClausesContext):
        limit = None
        if ctx.limitClause():
            limit = int(ctx.limitClause().INTEGER().getText())
        return limit


del SparqlParser
