//
// CONFIGURATION VARIABLES
//
// High and very query times will be emphasized in (classes .high and .veryhigh
// in the CSS).
var high_query_time_ms = 200;
var very_high_query_time_ms = 1000;
// Caching files is faster, but makes debugging harder.
var cache_files = false;
// Threshold beyond which a position in the suggestion list counts as infinite.
var infinite_position = 1000000;
// Page size for computing MRR (the parameter k from the paper).
var mrr_page_size = 7;
var ks_page_size = 7;
var dir_regex = "";  // "hannah";  // 2020-08-13.hannah"; 
var result_regex = "";

// 
// A FEW GLOBAL VARIABLES
//
// It's good and meaningful to know the base path of the URL and the currently
// selected output directory (folder with results) globally.
//
var base_path = window.location.pathname.replace(/www\/.*/, "results/");
console.log("Base Path: " + base_path);
var results_directory = "";

//
// THE ENTRY POINT FOR THE JAVASCRIPT CODE
//
// All the code for reading files and computing statistics is now nicely
// encapsulated in several classes below.
//
$(document).ready(function() {

  // Initialization
  $.ajaxSetup({ cache: cache_files, ifModified: true });
  $("div#detailsbox").hide();

  // Click on README link should open dialog box which renders MD as HTML. The
  // preventDefault prevents that the link is opened as it normally would in a
  // browser. The time is added to prevent caching.
  $("a.readme").click(function(e) {
    e.preventDefault();
    $.get($(this).attr("href") + "?_=" + new Date().getTime(), function(markdown) {
      var md2html = new showdown.Converter();
      md2html.setFlavor("github");
      console.log(showdown.getDefaultOptions());
      $("div.readme").html(md2html.makeHtml(markdown));
      $("div.readme a").attr("target", "_blank");
      $("div.readme").dialog({ width: 1000, height: "auto" });
    });
  });

  // Parse query parameters and set the corresponding variables. Set default
  // values at the beginning.
  if (window.location.search.startsWith("?")) {
    params = {}
    window.location.search.substring(1).split("&").forEach(function(pair_string) {
      pair = pair_string.split("=");
      if (pair.length == 2) { params[pair[0]] = pair[1]; }
    })
    if ("dir" in params) { dir_regex = params["dir"]; }
    if ("res" in params) { result_regex = params["res"]; }
    if ("mrr_page_size" in params) { mrr_page_size = params["mrr_page_size"]; }
    if ("ks_page_size" in params) { ks_page_size = params["ks_page_size"]; }
  }

  // Get available output directories. This works by parsing the directory
  // listing provided by Python's http.server
  $.get(base_path, function(data) {
    var output_directories = Array()
    $(data).find("a").each(function(){
      var name = $(this).attr("href");
      if (name.match(dir_regex)) {
        output_directories.push(name);
        $("select#directory").append(new Option(name, name));
      }
    });
    output_directories.sort();
    output_directories.reverse();
    $("select#directory").val(output_directories[0]);
    $("select#directory").trigger("change");
  });

  // Get all results from results directory. If result_regex is not the empty
  // string, filter by that
  $("select#directory").change(function(){
    // Remove details from previous output dir (confusing otherwise). Reset
    // filter regex. Remove previous tables.
    $("#details").attr("src", "about:blank");
    $("input#result-regex").val("");
    $("div.overview table").remove();
    // Get extract links from result directory. NOTE: this assumes a certain
    // format. It works when the pages are served with python3 -m http.server
    results_directory = $("select#directory option:selected").val();
    console.log("Result dir: " + results_directory);
    var result_names = new Array();
    $.get(base_path + results_directory, function(data) {
      $(data).find("a").each(function() {
        var filename = $(this).attr("href");
        if (filename.match(/\.completion\.results\.tsv$/) &&
              filename.match(result_regex)) {
          var result_name = filename.replace(
                              /\.completion\.results\.tsv$/, "");
          result_names.push(result_name);
        }
      });
      // Sort names using ResultNameSorter below.
      ResultNameSorter.sort(result_names);

      // Create the result tables using the class below (all the code is in the
      // constructor of that class).
      var result_tables = new ResultTables(result_names);
    })
  });

  // Reload results on click on the SVG icon and change color on mouseover.
  $("svg.reload").click(function() {
    console.clear();
    console.log("Reload button clicked");
    $("select#directory").trigger("change");
  });
  $("svg.reload").hover(
    function() { $("#svgbar").css({fill: "#000000"}); },
    function() { $("#svgbar").css({fill: "#aaaaaa"}); },
  );

  // Reload and recompute when different certificate analysis selected.
  $("select#certificate-analysis").change(function() {
    console.clear();
    console.log("Certificate analysis: " + $("#certificate-analysis").val());
    $("select#directory").trigger("change");
  });
});

// Class for sorting result names using the following criteria in this order:
//
// 1. fbeasy < freebase < wikidata
// 2. unranked < agnostic < sensitive < mixed
// 3. blazegraph < virtuoso < qlever
// 4. alphabetical
//
class ResultNameSorter {

  // Sort result names using the order described above (replaces result_names).
  static sort(result_names) {
    result_names = result_names.sort(this.compare);
  }

  // Compare two names using the criteria above in that order.
  static compare(name_1, name_2) {
    var rns = ResultNameSorter;
    return (rns.kb_key(name_1) - rns.kb_key(name_2))
             || (rns.mode_key(name_1) - rns.mode_key(name_2))
             || (rns.backend_key(name_1) - rns.backend_key(name_2))
             || name_1 < name_2;
  }

  // Return knowledge base key (taking the name does what we want).
  static kb_key(name) {
    if (name.match("fbeasy")) return 1;
    else if (name.match("freebase")) return 2;
    else if (name.match("wikidata")) return 3;
    else return 4;
  }

  // Return mode key.
  static mode_key(name) {
    if (name.match("unranked")) return 1;
    else if (name.match("agnostic")) return 2;
    else if (name.match("sensitive")) return 3;
    else if (name.match("mixed")) return 4;
    else return 5;
  }

  // Return backend key (take qlever as default, in case it's not specified).
  static backend_key(name) {
    if (name.match("blazegraph")) return 1;
    else if (name.match("virtuoso")) return 2;
    else return 3;
  }
}

// Create three result tables and fill them ("with context", "without context",
// "all tokens"). The procedure is as follows:
//
// 1. Create the table skeletons (using the class ResultTable below).
// 2. Show/hide tables and filter rows according to user selection.
// 3. Compute results asynchronously and update the respective table rows as
//    soon as a result has been computed.
//
// @3: Alternatively, we could first load all results and compute all staistics
// and only then draw the tables. That would be simpler, but then we would not
// have the nice effect of seeing the table grow as one row after the other
// becomes available.
class ResultTables {

  // All the code is in the constructor, so we could have just as well written
  // this as a global function.
  constructor(result_names) {

    // Create table skeletons (with empty rows, but already ids and actions
    // associated with it). The table will be filled below, when the results are
    // computed.
    var table_1 = new ResultTable(result_names, "with-context", "Results WITH context");
    var table_2 = new ResultTable(result_names, "wout-context", "Results WOUT context");
    var table_3 = new ResultTable(result_names, "all-tokens", "Results for ALL tokens");

    // Show only those tables for which the checkboxes are checked and react to
    // changes.
    $("span#table-selection input:checkbox").change(function() {
      var input_class = $(this).attr("class");
      if ($(this).is(":checked")) {
        $("div." + input_class + " table").show();
      } else {
        $("div." + input_class + " table").hide();
      }
    })
    $("span#table-selection input:checkbox").trigger("change");

    // Filter results by regex in input field #result-regex. Note that
    // String.search does a regex search, which is great.
    $("input#result-filter").focus();
    $("input#result-filter").keyup(function() {
      var filter_keywords = $(this).val().split(/\s+/);
      $("div.overview tbody tr").each(function() {
        var name = $(this).children(":first-child").text();
        var show_row =
          filter_keywords.every(keyword => name.search(keyword) != -1);
        if (show_row) $(this).show(); else $(this).hide();
      });
    });

    // Compute all the results and fill the associated table rows.
    var results = new Array();
    result_names.forEach(function(result_name, i) {
      var result = new Result(result_name);
      results[i] = result;
      result.read_from_file().
        then(function() {
          result.compute_statistics();
          console.log("Done: " + result.name);
          table_1.update_table_row(result, i);
          table_2.update_table_row(result, i);
          table_3.update_table_row(result, i);
        });
    });

    // Check if we should use query times or client times. If change, update
    // table rows.
    $("input:checkbox.use-client-time").change(function() {
      use_client_time_instead_of_query_time = $(this).is(":checked");
      results.forEach(function(result, i) {
        var rows_already_read = result.rows;
        result = new Result(result_names[i]);
        result.rows = rows_already_read;
        result.compute_statistics();
        console.log("Recomputed: " + result.name);
        table_1.update_table_row(result, i);
        table_2.update_table_row(result, i);
        table_3.update_table_row(result, i);
      });
    })
  }
}

// Class for a single result table. The constructor creates a skeleton (with
// empty rows) and the function update_table_row can be used to fill a table
// row, as soon as the respective result has been loaded and processed.
class ResultTable {

  // Create single table and rememver array of table row ids. A click on a table row
  // will lead to the corresponding detailed result.
  //
  // NOTE: We need to remember the table row ids in the order of the result
  // names. That is important because in ResultTables above, the lines are NOT
  // filled in order.
  constructor(result_names, table_name, table_description) {
    this.name = table_name;
    // Table headers.
    var headers_with_explanation = new Array(
      [table_description,
         "The names in this column are the base names of the files in the " +
         "results directory.<br>A click on this header computes LaTeX code " + 
         "for this table and copies it to your clipboard."],
      ["Tokens",
        "The total number of tokens completed (depending on the experiment, " +
        "this may or may not include tokens, which are variables)."],
      ["failed",
        "The number of tokens, for which ALL AC queries failed."],
      ["Queries",
        "The total number of AC queries executed."],
      ["failed",
        "The number of AC queries that failed, for each prefix length."],
      // "Median",
      // "Mean",
      ["≤ 0.2s",
        "The percentage of tokens which took ≤ 0.2 seconds. For each " +
        " token, the <b>maximum</b> time per AC query is taken."],
      ["≤ 1s",
        "The percentage of tokens which took ≤ 1.0 seconds. For each " +
        "token, the <b>maximum</b> time per AC query is taken."],
      ["Max",
        "The maximum time of a AC query in milliseconds."],
      ["Sens",
        "Percentage of sensitive queries that did not fail. " +
        "For &quot;agnostic&quot;, this is 0% by definition. "],
      ["MRR<sub>" + mrr_page_size + "</sub>",
        "The Mean Reciprocal Rank (MRR) for each prefix length, where the subscript " +
        " says how many results fit on one page.<br>For one AC query, the " +
        "reciprocal rank is computed as follows. The rank is r = ceil(i/k) " +
        "where i is the position of the token (starting at 1) in the result list. " +
        "That is, if i <= k (that is, the token is on the first page of the " +
        "result, assuming a page size of k), then the rank is 1. The reciprocal " +
        "rank is just the inverse 1/r."],
      // "MRR<sub>" + mrr_page_size + "</sub> &nbsp; histogram @ 0",
      ["KS<sub>" + ks_page_size + "</sub>",
        "The average KS (number of keystrokes) required per token, where the index " +
        "says how many results fit in one page.<br>For each token, " +
        "KS is the first prefix length (among those tried out in the evaluation), " +
        "for which the desired token is on the first page of results. " +
        "When this does not hold for any of the prefix lengths, KS is taken as the " +
        "length of the token plus 1 (modeling that the token had to be typed " +
        "completely, including an end-of-token character to signal that this is " +
        "not a prefix)."]
    );
    var table_headers = "";
    headers_with_explanation.forEach(function(header_with_explanation) {
      table_headers += "<td title=\"" + header_with_explanation[1] + "\">"
                         + header_with_explanation[0]
                         + "</td>";
    });
    // Table with header and empty body.
    var table =
      "<table class=\"" + table_name + "\" class=\"tablesorter\">" +
      "<thead><tr>" + table_headers + "</tr></thead>" +
      "<tbody></tbody>" +
      "</table>";
    $("div." + table_name).html(table);

    // With click on first thead, produce LaTeX source code of table.

    var _this = this;
    $("div." + table_name + " thead td:first-child").click(function() {
      _this.produce_latex();
    });

    // Nicer formatting of tooltips. Setting "track" to true lets the tooltip
    // follow the mouse. Setting content to the specified function allows HTML
    // in the tooltip (default: plain text).
    $(document).tooltip({
      track: true,
      content: function () { return $(this).prop('title'); }
    });

    // Now add the empty table rows. The id for each table row is a
    // concatenation of the table name and the results name, where dots in the
    // result name are replaced by -. A click on a row should trigger loading
    // detail information about that result in the iframe#details.
    var table_row_ids = new Array();
    result_names.forEach(function(result_name, i) {
      // A CSS id must not have dots (this cost me some time to find out).
      // Highligh the rows with mixed qlever (see CSS for the "best" class)
      var result_name_id = table_name + "-" + result_name.replace(/\./g, "-");
      var table_class = "";
      if (result_name.match("qlever")) {
        if (result_name.match("sensitive")) table_class = " class=\"best1\"";
        else if (result_name.match("mixed")) table_class = " class=\"best2\"";
      }
      var table_row = "<tr id=\"" + result_name_id + "\"" + table_class + "></tr>";
      table_row_ids.push(result_name_id);
      $("div." + table_name + " tbody").append(table_row);
      $("tr#" + result_name_id).click(function() {
        // Highlight the table row. Make sure that any table rows that were
        // highlighted before are not highlighted anymore.
        $("div.overview tbody tr").removeClass("selected");
        $("tr#" + result_name_id).addClass("selected");
        // Load detail view for this result into #details iframe.
        var relative_url =
          base_path.replace(/results.*/, "www/details.html") +
          "?dir=" + results_directory.replace(/\/$/, "") +
          "&res=" + result_name;
        console.log("Show details for: " + relative_url);
        $("iframe#details").attr("src", relative_url);
        $("div#detailsbox").show();
      });
    });
    this.table_row_ids = table_row_ids;;

    // NEW: Also create an array with the rows of a corresponding latex table.
    this.latex_rows = new Array(table_row_ids.length);
  }

  // Update the table row with the given index using the given result.
  //
  // NEW: Also update the corresponding latex table row.
  update_table_row(result, i) {
    // Different statistics depending on the table.
    var qt, mrr, ks;
    switch (this.name) {
      case "all-tokens":
        qt = result.qt_all;
        mrr = result.mrr_all;
        ks = result.ks_all;
        break;
      case "with-context":
        qt = result.qt_with_context;
        mrr = result.mrr_with_context;
        ks = result.ks_with_context;
        break;
      case "wout-context":
        qt = result.qt_wout_context;
        mrr = result.mrr_wout_context;
        ks = result.ks_wout_context;
        if (qt.num_queries == 0) {
          return;
        }
        break;
    }



    // Values for the table row in an array.
    var values = [result.name.replace("timeout", "to").replace("certifyTrue", "certify"),
                  qt.num_tokens,
                  qt.num_tokens_failed_completely,
                  qt.num_queries,
                  qt.num_queries_failed_string,
                  // qt.median + " ms",
                  // qt.mean + " ms" ,
                  qt.perc_200ms + " %",
                  qt.perc_1000ms + " %",
                  qt.max + " ms",
                  qt.perc_queries_sensitive + " %",
                  mrr.averages_string,
                  // "<small>" + mrr.histogram_string + "</small>",
                  ks.avg.toFixed(2)];

    // Surround each value by <td>...</td> and concatenate. Then update the
    // table row. Note that a table row selection does not include the <tr> and
    // </tr> tags, just what's within.
    var table_row = values.map(value => "<td>" + value + "</td>").join("");
    $("#" + this.table_row_ids[i]).html(table_row);

    // HACK: Estimate timeout (seconds) from maximum query time (milliseconds).
    var timeout_seconds = qt.max < 2000 ? 1 : 5;

    // Percentage of failed queries. If at least 1%, round to integer (e.g. 17).
    // If below 1% show with on digit after the dot (e.g. 0.4).
    var perc_queries_failed = 100 * qt.num_queries_failed / qt.num_queries;
    perc_queries_failed = Math.round(perc_queries_failed) > 0
      ? Math.round(perc_queries_failed).toFixed(0)
      : perc_queries_failed.toFixed(1);

    // Row in latex table as one flat array.
    //
    // (Note that qt.num_queries_failed_string and mrr.averages_array are
    // themselves arrays of arrays. The flat(2) call flattens to a depth of
    // two.)
    this.latex_rows[i] =
      [result.name,
       // String(qt.num_tokens),
       // "$" + Math.round(100 * qt.num_tokens_failed_completely /
       //   qt.num_tokens) + "\\p$",
       // qt.num_tokens_failed_completely
       // String(qt.num_queries),
       // qt.num_queries_failed == 0
       //   ? ["", "", "", "", "", 0]
       //   : qt.perc_queries_failed_array.map(x => [x[0] + ":", "$" + x[1] + "\\p$"]),
       "$" + qt.perc_200ms + "\\p$",
       "$" + qt.perc_1000ms + "\\p$",
       qt.num_queries_failed == 0
         ? "\\multicolumn{6}{r||}{$" + qt.max + "\\ms$}"
         // NEW 2021-01-19: Report the percentage of AC queries that failed and
         // not the percentage of tokens that failed completely.
         : "\\multicolumn{6}{r||}{$" + perc_queries_failed + "\\p > " + timeout_seconds + "\\s$}",
         // : "\\multicolumn{6}{r||}{$" + Math.round(100 * qt.num_tokens_failed_completely /
         //      qt.num_tokens) + "\\p > " + timeout_seconds + "\\s$}",
       // "$" + qt.max + "\\ms$",
       result.name.match(/(unranked|agnostic)/)
         ? "$0\\p$"
         // : "$100\\p$",
         : "$" + Math.round(100 - 100 * (qt.num_queries_agnostic + qt.num_queries_failed)
                              / qt.num_queries) + "\\p$",
       mrr.averages_array.map(x => [x[0] + ":", "$" + x[1] + "\\p$"]),
       String(ks.avg.toFixed(2))].flat(2);
    // console.log(this.latex_rows[i]);
  }

  // Produce LaTeX source code for this table and copy it to the clipboard.
  produce_latex() {
    console.log("Produce LaTeX for table: " + this.name);
    // console.log("Produce from this array of size "
    //   + this.latex_rows.length + ": " + this.latex_rows);
    var cols = 17; // Number of columns in the LaTex table.
    var kb_ids = ["fbeasy", "freebase", "wikidata"];

    // HACK: hardcode number of tokens based on date of evaluation
    var is_new = results_directory.startsWith("2021")
    var num_tokens_fbeasy = is_new ? 314 : 292;
    var num_tokens_freebase = is_new ? 478 : 443;
    var num_tokens_wikidata = is_new ? 1258 : 1031;

    var kb_headings = [
      "Fbeasy (" + num_tokens_fbeasy + " tokens)",
      "Freebase (" + num_tokens_freebase + " tokens)",
      "Wikidata (" + num_tokens_wikidata + " tokens)"];
      // (362M triples)", //  (1.9B triples)", // (6.9B triples)"];
    var latex = [];
    // Comment that clarifies the origin of this code.
    latex.push("% Copied from " + window.location.href +
               " on " + new Date().toLocaleString());
    latex.push("");
    // Begin tabular.
    latex.push(
        ["\\setlength{\\tabcolsep}{2.5mm}",
         "\\def\\spa{\\hspace{0.7mm}}",
         "\\def\\spb{\\hspace{2.0mm}}",
         "\\def\\b#1{{\\color{darkblue} #1}}",
         "\\def\\bb#1{{\\color{darkblue}\\boldmath\\textbf{#1}}}",
         "{\\renewcommand{\\baselinestretch}{1.2}\\normalsize",
         "\\begin{tabular}{|l|l||r|r|r@{\\spa}r@{\\spb}r@{\\spa}r@{\\spb}r@{\\spa}r||"
          + "r@{\\spa}|r@{\\spa}r@{\\spb}r@{\\spa}r@{\\spb}r@{\\spa}r|r|}",
         ""].join("\n"));
    // Iterate over all KBs.
    for (var i = 0; i < kb_ids.length; i++) {
      var kb_id = kb_ids[i];
      var kb_heading = kb_headings[i];
      // Header of table for that KB.
      if (i > 0) latex.push("\\multicolumn{" + cols + "}{l}{} \\\\[-2mm]");
      latex.push("\\hline");
      // latex.push("\\multicolumn{" + cols + "}{l}{" + kb_heading + "} \\\\\\hline");
      latex.push(
        ["\\multicolumn{2}{|l||}{\\textbf{" + kb_heading + "}}",
         // "\\textbf{Tokens}",
         // "\\textbf{failed}",
         // "\\textbf{Queries}",
         // "\\multicolumn{6}{r||}{\\makebox{\\bf failed}}",
         "$\\mathbf{\\le 0.2\\s}$",
         "$\\mathbf{\\le 1.0\\s}$",
         "\\multicolumn{6}{r||}{\\boldmath\\textbf{Max}}",
         // "\\makebox{$\\mathbf{max}$}",
         "$\\mathbf{Sens}$",
         "\\multicolumn{6}{r|}{\\boldmath\\textbf{MRR$_7$}}",
         "\\boldmath\\textbf{KS$_7$}"]
        .join(" & ") + " \\\\\\hline");

      // Iterate over all rows in table for that KB. We are assuming that the
      // first cell of each row has the form <kb>.<mode>.<prefixes>.<engine>
      this.latex_rows.forEach(function(latex_row, i) {
        var result_name = latex_row[0];

        // Skip certain results.
        if (result_name.match(/timeout/)
             || result_name.match(/certify/)
             || result_name.match(/0-4-7/)) {
          console.log("Filtered out LATEX row: " + result_name);
          return;
        }
        
        // Parse mode and engine from remaining result name. A typical name is 
        // wikidata.sensitive.0-3-7.qlever
        var result_name_parts = latex_row[0].split(".");
        if (result_name_parts[0].startsWith(kb_id)) {
          var mode = result_name_parts[1];
          mode = mode.charAt(0).toUpperCase() + mode.slice(1);
          var engine = result_name_parts[3];
          engine = engine.charAt(0).toUpperCase() + engine.slice(1);
          var engine_abbrev = engine.charAt(0).toUpperCase();
          var modified_latex_row = [mode, engine].concat(
            latex_row.slice(1));

          // Emphasize Qlever row (sensitive in blue and blold, mixed in blue).
          if (engine_abbrev == "Q") {
            var emph = "";
            if (mode == "Sensitive") emph = "\\b";
            if (mode == "Mixed") emph = "\\bb";
            if (emph != "") {
              modified_latex_row =
                modified_latex_row.map(x => 
                  x.startsWith("\\multicolumn")
                    ? x.replace(
                        /^(\\multicolumn{[^}]+}{[^}]+}){(.*)}$/,
                        "$1{" + emph + "{$2}}")
                    : emph + "{" + x + "}"); 
            }
          }
          // Horizontal line before first Sensitive and first Mixed (which is
          // Blazegraph).
          if (engine_abbrev == "B") latex.push("\\hline");

          latex.push(modified_latex_row.join(" & ") + " \\\\");
        }
      });
      latex.push("\\hline");
      latex.push("");
    }
    // END tabular.
    latex.push("\\end{tabular}}");

    // Join lines, copy to textarea and from there to the clipboard. The
    // textarea is briefly displayed on the screen.
    var latex_text = latex.join("\n");
    // console.log(latex_text);
    $("div.latex").show();
    $("div.latex textarea").html(latex_text);
    $("div.latex textarea").select();
    document.execCommand("copy");
    var show_duration_seconds = 5;
    console.log("COPIED TO CLIPBOARD, showing text area for "
      + show_duration_seconds + " seconds");
    setTimeout(function() { $("div.latex").hide(); }, show_duration_seconds * 1000);
  }
}
