//
// CONFIGURATION VARIABLES
//
// High and very query times will be emphasized in (classes .high and .veryhigh
// in the CSS).
var high_query_time_ms = 200;
var very_high_query_time_ms = 1000;
// Caching files is faster, but makes debugging harder.
var cache_files = true;
// Threshold beyond which a position in the suggestion list counts as infinite.
var infinite_position = 1000000;
// Page size for computing MRR (the parameter k from the paper).
// var mrr_page_size = 7;
var dir_regex = "";  // "hannah";  // 2020-08-13.hannah"; 
var result_regex = "";
var use_client_time_instead_of_query_time = false;

// Consider query FAILED if certificate failed.
// var certificate_analysis = "target certificate";
// var certificate_analysis = "page certificate";
// var certify_analysis = "";

// This file provides the following classes:
//
// class ResultRow:    values from one row of one results.tsv file
// class Result:       all values from one results.tsv file + statistics
// class ResultTable:  overview table with all results from a directory
//
// class QtStatistics:    statistics for query time (QT)
// class MrrStatistics:   statistics for mean reciprocal rank (MRR)
// class KsStatistics:    statistics for number of keystrokes (KS)
//


// Class for one row from the results file.
class ResultRow {

  // Parse row values from given tab-separated values. If something changes in
  // the output format, change it here.
  constructor(line) {
    var fields = line.split("\t");

    // Parse the various fields depending on their type, with a little pimping
    // here and there.
    this.target_query_name     = fields[0];
    this.completion_query_name = fields[1];
    this.partial_triple        = fields[3].replace(/\s*$/, "");
    this.alias                 = fields[4];
    this.triple_num            = parseInt(fields[5]);
    this.has_context           = parseInt(fields[6]);
    this.query_time            = fields[7] != "" ? parseInt(fields[7]) : Number.MAX_VALUE;
    this.client_time           = fields[11] != "" ? parseInt(fields[11]) : Number.MAX_VALUE;
    // if (this.query_time == Number.MAX_VALUE) 
    //   console.log("Failed query encountered: " + this.completion_query_name);
    // Note: infinite_position is what the evaluation scripts writes when the
    // desired entity is not in the top-k suggestions.
    var infinite_position = 1000000;
    this.position              = parseInt(fields[9]);
    if (this.position >= infinite_position) this.position = Number.MAX_VALUE;
    this.prefix_length         = parseInt(fields[10]);
    this.client_time           = fields[11] != "" ? parseInt(fields[11]) : Number.MAX_VALUE;
    this.is_warmup_query       = parseInt(fields[12]) == 1;
    this.is_new_token          = parseInt(fields[13]) == 1;
    this.is_very_old_eval      = fields.length <= 14;
    // This column is now redundant, but we keep it for a while, for backwars
    // compatibility with earlier results.
    // this.is_fallback           = fields.length > 14 ? parseInt(fields[14]) : 0;
    this.is_mixed_sensitive    = fields.length > 14 ? fields[14] == "sensitive" : false;
    this.is_mixed_agnostic     = fields.length > 14 ? fields[14] == "agnostic" : false;
    this.is_single             = !this.is_mixed_sensitive && !this.is_mixed_agnostic;
    this.is_variable_query     = fields.length > 15 ? parseInt(fields[15]) : 0;
    this.is_old_eval           = fields.length <= 16;
    this.has_failed            = fields.length > 16 ? parseInt(fields[16]) == 1
                                                    : this.query_time == Number.MAX_VALUE;
    this.is_last_of_token      = fields.length > 17 ? parseInt(fields[17]) : 0;
    this.target_cert_failed    = fields.length > 18 ? parseInt(fields[18]) : 0;
    this.page_cert_failed      = fields.length > 19 ? parseInt(fields[19]) : 0;

    // HACK: If is_certify_run, take !target_certified or !first_page_certified
    // as failed
    // if (this.is_certify_run) this.has_failed = this.first_page_certified == 0;

    // HACK: simulate a smaller timeout.
    // var smaller_timeout = 1000;
    // if (this.query_time > smaller_timeout) {
    //   this.query_time = smaller_timeout;
    //   this.position = Number.MAX_VALUE;
    //   this.has_failed = true;
    // }

  }
}

// Compute statistics for given array of query times (mean, median, max, etc.)
//
// It works as follows. Creat object with constructor, add query times with
// add_query_time, and then compute statistics with compute_statistics.
//
// if the query time is Number.MAX_VALUE, the query is considered to have
// failed.
//
// The number of failed queries is considered for the median and for the % of
// queries faster than 200ms and 1000ms, respectively. If half or more of the
// queries have failed, the median is ∞ (Unicode 221E). 
// 
// NOTE: this.num_all and this.num_failed count different things depending on
// the aggregation mode (queries for "all", tokens for "max" and "sum").
//
// NEW: Like for MrrStatistics, count the number of failed tokens per prefix
// length.
class QtStatistics {

  // Create empty object.
  constructor(aggregation_mode) {
    this.aggregation_mode = aggregation_mode;
    this.query_times = [];
    this.token_has_failed_completely = [];
    this.num_tokens = 0;
    this.num_queries = 0;
    this.num_tokens_failed_once = 0;
    this.num_tokens_failed_completely = 0;
    this.num_queries_failed = 0;
    this.num_queries_failed_per_prefix_length = [0];
    this.num_queries_sensitive = 0;
  }

  // Add query time in given aggregation mode.
  //
  // NEW: we now explicitly have the information that a query has failed and
  // we need (and must) no longer derive it from query_time == Number.MAX_VALUE
  update(query_time, has_failed, is_new_token, prefix_length) {
    // Make sure that we have a counter for each prefix length that we have seen
    // at least once.
    if (!(prefix_length in this.num_queries_failed_per_prefix_length))
      this.num_queries_failed_per_prefix_length[prefix_length] = 0;

    // Keep track of the number of queries.
    this.num_queries += 1;

    // Keep track of the number of tokens and for each token keep track of
    // whether it has failed once and whether it has succeeded once already.
    //
    // In the end, we also need to know whether the token has failed
    // completeley. To achieve this, we set token_has_failed_completely to 1
    // when we see a new token and set it to 0 as soon as one query for that
    // token fails.
    if (is_new_token) {
      this.num_tokens += 1;
      this.has_current_token_failed_yet = false;
      this.seen_non_failed_query_in_token = false;
      this.token_has_failed_completely.push(1);
    }

    // Keep track of the number of failed queries (overall and per prefix
    // length) and of the number of tokens that failed once.
    if (has_failed) {
      this.num_queries_failed += 1;
      this.num_queries_failed_per_prefix_length[prefix_length] += 1;
      if (!this.has_current_token_failed_yet) {
        this.num_tokens_failed_once += 1;
        this.has_current_token_failed_yet = true;
      }
    }

    // Call the proper function depending on the mode (only if the query has not
    // failed) and record that we have seen a non-failed query for this token.
    else {
      if (this.aggregation_mode == "all") this.update_mode_all(query_time);
      else if (this.aggregation_mode == "max") this.update_mode_max(query_time);
      else if (this.aggregation_mode == "sum") this.update_mode_sum(query_time);
      else console.error("QtStatistics.update: Invalid aggregation mode");
      this.seen_non_failed_query_in_token = true;
      console.assert(this.token_has_failed_completely.length = this.num_tokens);
      this.token_has_failed_completely[this.num_tokens - 1] = 0;
    }
  }

  // Mode ALL: add all (valid, see above) query times of a token.
  update_mode_all(query_time) {
    this.query_times.push(query_time);
  }

  // Mode MAX: take the max of all (valid, see above) query times of a token.
  update_mode_max(query_time) {
    if (!this.seen_non_failed_query_in_token) this.query_times.push(0);
    var i = this.query_times.length - 1;
    this.query_times[i] = Math.max(this.query_times[i], query_time);
  }

  // Mode SUM: take the sum of all (valid, see above) query times of a token.
  update_mode_sum(query_time) {
    if (!this.seen_non_failed_query_in_token) this.query_times.push(0);
    var i = this.query_times.length - 1;
    this.query_times[i] += query_time;
  }

  // Compute statistics for the query times collected thus far, as described in
  // the comment for the class above.
  //
  // TODO: think about deleting the query times at this point, since they take
  // up space and are no longer needed. But maybe this doesn't matter.
  compute_statistics() {
      if (this.num_queries == 0) {
        return;
      }
    // We have three variants of #failed:
    //
    // 1. The number of tokens, for which ALL queries failed.
    // 2. The number of tokens, for which ONE query failed
    // 3. The number of queries which failed.
    //
    // We currently take variant 1. Note that for MAX and SUM, there will be one
    // query per token unless all queries failed for that token.
    var n = this.query_times.length;
    this.num_tokens_failed_completely =
      this.token_has_failed_completely.reduce((prefix_sum, x) => prefix_sum + x);
    var N = this.aggregation_mode == "all" ? this.num_queries : this.num_tokens;
    // this.num_failed = this._mode == "all"
    //   ? this._num_failed_queries : this._num_failed_tokens;

    // Extract non-empty elements of the array. Gives an array of pairs, e.g.
    // [[0, 25], [3, 17], [7, 5]], where the first element of each pair is the
    // prefix length and the second is the number of failed queries for that
    // prefix length.
    this.num_queries_failed_array =
      this.num_queries_failed_per_prefix_length
          .map((x, i) => [i, x])
          .filter(x => x);
    this.perc_queries_failed_array =
      this.num_queries_failed_per_prefix_length
          .map((x, i) => [i, Math.round(100 * x / this.num_tokens)])
          .filter(x => x);
    
    // Percentage of mixed queries, where agnostic was used.
    this.perc_queries_sensitive =
      Math.round(100 * this.num_queries_sensitive / this.num_queries);
      // this.num_agnostic;

    // The same with HTML formatting for our table and some backwards
    // compatibility with older experiments.
    this.num_queries_failed_string =
      this.num_queries_failed_per_prefix_length.filter(x => x != undefined).length == 1
        ? this.num_queries_failed_per_prefix_length.filter(x => x != undefined)
        : this.num_queries_failed_per_prefix_length
           .map((x, i) => i + ":" + x + "</span>")
           // .map((x, i) => i + ":<span class=\"num\">" + x + "</span>")
           .filter(x => x != undefined)
           .join(", ");
           // .join(",&nbsp;&nbsp;&nbsp;");
    if (this.num_queries_failed == 0) this.num_queries_failed_string = "0";
    // if (this._mode == "max") console.log("qt = " + this.query_times.length + ", all = " + this.num_all);
    // this.num_all = this.query_times.length + this.num_failed;
    // Compute statistics.
    this.query_times = this.query_times.sort((x, y) => x - y);
    this.mean = this.query_times.length == 0 ? 0 
                  : this.query_times.reduce((prefix_sum, x) => prefix_sum + x) / n;
    var median_index = Math.round(0.5 * N);
    this.median = median_index < n ? Math.round(this.query_times[median_index]) : "\u221E";
    this.max = this.query_times.length == 0 ? "?" : this.query_times[n - 1];
    this.perc_200ms = this.query_times.filter(x => x <= high_query_time_ms).length / N;
    this.perc_1000ms = this.query_times.filter(x => x <= very_high_query_time_ms).length / N;
    // Some rounding etc.
    this.mean = Math.round(this.mean);
    this.perc_200ms = Math.round(100 * this.perc_200ms);
    this.perc_1000ms = Math.round(100 * this.perc_1000ms);
  }
}

// Class for MRR statistics (Mean Reciprocal Rank per token).
//
// NEW: Now one histogram per prefix length. We simply use an Array here. That
// way, we can be sure to have the histograms in order. And JS Arrays have the
// nice property that when you write to an index beyond the current length of
// the Array, the length of the Array is simply extended. In fact JS Arrays are
// not "real arrays", but "associative arrays" aka maps.
class MrrStatistics {

  // Create empty statistics. Dictionaries for prefix lengths > 0 are added as
  // we go along, see function update below.
  constructor() {
    this.histograms = [{}];
    this.sum = [0];
    this.num_tokens = [0];
    this.avg = [];
    this.max_prefix_length_seen = 0;
  }

  // Add new position. Store infinite position as ∞ (Unicode 221E).
  // The reciprocal rank for the infinite position is simply 0.
  update(position, prefix_length) {
    // Abbreviate the prefix length to i for better readability of the code.
    var i = prefix_length

    // The max_prefix_length_seen is for backwards compatibility with
    // evaluations with a fixed prefix length k, but variable prefix lengths in
    // the result because some tokens have less than k letters.
    // if (i > this.max_prefix_length_seen) this.max_prefix_length_seen = i;

    // If we see this prefix length for the first time, add a new histogram and
    // a new num_tokens count.
    if (!(i in this.histograms)) {
      this.histograms[i] = {};
      this.num_tokens[i] = 0;
      this.sum[i] = 0;
    }

    // Update statistics for that prefix length.
    var rank = position < Number.MAX_VALUE ? 1 + Math.floor((position - 1) / mrr_page_size) : "\u221E";
    var reciprocal_rank = position < Number.MAX_VALUE ? 1 / rank : 0;
    this.histograms[i][rank] = (this.histograms[i][rank] || 0) + 1;
    this.sum[i] += reciprocal_rank;
    this.num_tokens[i] += 1;
  }

  // Compute (additional) statistics.
  //
  compute_statistics() {
    // Compute the average for each token length. Note that this.avg is still
    // empty at this point, so we iterate over this.sum
    var _this = this;
    this.sum.forEach(function(value, i) {
      if (_this.num_tokens[i] > 0) 
        _this.avg[i] = Math.round(100 * _this.sum[i] / _this.num_tokens[i]);
    });
    // Extract non-empty averages. Gives something like this: [[0, 17], [3, 23],
    // [7, 15]], where the first element of a subarray is the prefix length and
    // the second is the MRR in percent computed i the previous forEach loop.
    this.averages_array = this.avg.map((x, i) => [i, x]).filter(x => x);
    // If we only have prefix length 0, show a single percentage without 0: in
    // front (because for older evaluations, the MRR may come from a single
    // prefix length > 0).
    // backwards compatibility with older results (would be shown as 0: ...
    // otherwise even if the single prefix length was > 0).
    var only_prefix_length_zero =
      this.avg.filter(x => x != undefined).length == 1
        && this.num_tokens[0] > 0;
    this.averages_string = only_prefix_length_zero
      ? this.avg[0] + "%"
      : this.avg.map((x, i) => i + ":<span class=\"perc\">" + x + " %</span>")
                .filter(x => x)
                .join(",&nbsp;&nbsp;&nbsp;");
    // NOTE: For now, the histogram string contains only the histogram for prefix
    // length 0. If that histogram is empty, all numbers are simply 0.
    var histogram_0 = this.histograms[0];
    var num_tokens_0 = this.num_tokens[0];
    Object.keys(histogram_0).forEach(function(rank) {
      histogram_0[rank] = Math.round(100 * histogram_0[rank] / num_tokens_0);
    })
    this.histogram_string =  "[" +
      "1: "      + (histogram_0[1] || 0) + " %, " +
      "2: "      + (histogram_0[2] || 0) + " %, ..., " +
      "\u221E: " + (histogram_0["\u221E"] || 0) + "%]";
  }

}

// Class for KS statistics (Number of keystrokes per token).
class KsStatistics {

  // Create empty statistics.
  constructor() {
    this.ks_per_token = [];
    this.num_tokens = 0;
    this.histogram = {};
    this.max = 0;
    this.sum = 0;
    this.avg = 0;
  }

  // Add new prefix length. For each token, compute the smallest prefix length
  // (among the prefix lengths evaluated), for which the token is in the top-k
  // results of the AC query, where k = mrr_page_size. If this happens for NONE
  // of the prefix length, take max_prefix_length (computed by the caller as
  // length of the token + 1), which corresponds to typing a token completely
  // (including the $ for exact match).
  update(prefix_length, max_prefix_length, is_new_token, position) {
    if (is_new_token) {
      this.ks_per_token.push(max_prefix_length);
      // this.ks_per_token.push(0);
      this.num_tokens += 1;
      this.found_token = false;
    }

    if (!this.found_token && position <= ks_page_size) {
      this.ks_per_token[this.ks_per_token.length - 1] =
        Math.min(prefix_length, max_prefix_length);
      this.found_token = true;
    }
  }

  // Compute the histogram.
  compute_statistics() {
    var _this = this;
    this.ks_per_token.forEach(function(ks) {
      _this.histogram[ks] = (_this.histogram[ks] || 0) + 1;
      if (ks > _this.max) _this.max = ks;
      _this.sum += ks;
   })
    this.avg = this.sum / this.num_tokens;
  }
}


// Class for processing result from <result_name>.completion.results.csv file.
// Usage:
//
// var result = new Result(result_name);
// result.read_from_file().
//   then(function() {
//     result.compute_statistics()
//     ...
//   });
//
class Result {

  // Create from given name.
  constructor(result_name) {
    // Result name of the form <knowledge base>.<mode>.<config>.<engine>
    // Extract mode and construct path.
    this.name = result_name;
    this.completion_mode = result_name.split(".")[1]
    this.path = base_path + results_directory +
                result_name + ".completion.results.tsv";
    // The raw data.
    this.rows = new Array();
    // The statistics.
    this.qt_all = new QtStatistics("max")
    this.qt_with_context = new QtStatistics("max")
    this.qt_wout_context = new QtStatistics("max")
    this.mrr_all = new MrrStatistics();
    this.mrr_with_context = new MrrStatistics();
    this.mrr_wout_context = new MrrStatistics();
    this.ks_all = new KsStatistics();
    this.ks_with_context = new KsStatistics();
    this.ks_wout_context = new KsStatistics();
  }

  // Read data from given file into this.rows, which is an array of objects of
  // type ResultRow, see above. Returns a Promise, see above about how to use.
  read_from_file() {
    var _this = this;
    return new Promise(function(resolve, reject) {
      // Get contents of results TSV, split by newline, and iterate over the lines
      // (expect the first, which are the headers, and the part after the final
      // newline, which is the empty string).
      //
      // NOTE: the "_this" is needed, so that it can be referred to in the callback
      // function of $.get (where "this" is overwritten by some jQuery object).
      // That methode is called "Closure".
      $.get(_this.path, function(data) {
        var lines = data.split("\n").slice(1, -1);
        var num_parse_errors = 0
        lines.forEach(function(line, i) {
          try {
            var result_row = new ResultRow(line);
            _this.rows.push(result_row);
          } catch (error) {
            console.error(
              "Error parsing row " + i + " from file " + 
              "\"" + _this.path + "\": " + error);
            num_parse_errors += 1;
            if (num_parse_errors >= 10) console.error("Too many parse errors");
            return num_parse_errors < 10;
          }
        });
        resolve();
      });
    });
  }

  // Compute statistics on query times, MRR and KS. Distinguish between tokens
  // with context and without context. Pay attention to failed queries and keep
  // track of how many there are (and how many tokens are affected).
  compute_statistics() {
    // Check that result is not empty.
    if (this.rows.length == 0) {
      console.error("Cannot compute statistics for empty result");
      return;
    }

    // Iterate over all rows of the result.
    //
    // NEW: For mixed completion mode there are now two result rows (with query
    // types "sensitive" and "agnostic") For the statistics, we consider the
    // first row is row.has_failed == 0 and otherwise the second row. For the
    // query time, if we take the second row, we form the maximum with the query
    // time of the first row (which is the timeout).
    var _this = this;
    var sensitive_has_failed;
    var sensitive_was_first_token;
    var sensitive_query_time;
    // DEBUG: percentage of rows with is_old_eval
    // console.log("Old eval rows: " + Math.round(
    //   100 * this.rows.filter(x => x.is_old_eval).length / this.rows.length) + "%");
    this.rows.forEach(function(row, i) {

      // NEW: check which query time we should use.
      var query_time = row.query_time;
      if (use_client_time_instead_of_query_time && query_time != Number.MAX_VALUE)
        query_time = row.client_time;

      // In MIXED completion mode, do the following: If this is the sensitive
      // query and it has failed, remember that and the query time and take the
      // maximum with agnostic in the next round.
      var is_new_token = row.is_new_token;
      if (row.is_mixed_sensitive) {
        sensitive_has_failed = row.has_failed;
        sensitive_query_time = query_time;
        sensitive_was_first_token = row.is_new_token;
      } else if (row.is_mixed_agnostic && sensitive_has_failed) {
        // TODO HACK: The blazegraph evaluation currently does not write the timeout
        // value here, but leaves the query time empty, so that it becomes
        // infinity. For now, hard-code a timeout of 1s.
        query_time = Math.max(query_time, sensitive_query_time);
        if (query_time = Number.MAX_VALUE) query_time = 1000
        is_new_token = sensitive_was_first_token;
      }

      // In mixed completion mode, consider this row only if either: (1) is is
      // sensitive and has not failed, or (2) it is agnostic and sensitive has
      // failed before.
      var consider_row =
        (row.is_single)
          || (row.is_mixed_sensitive && !row.has_failed)
          || (row.is_mixed_agnostic && sensitive_has_failed);

      // NEW: If certify analysis, consider failed if certification failed (there
      // are two ways how a certification can fail).
      var has_failed = row.has_failed;
      if (row.is_mixed_agnostic && sensitive_has_failed) {
        if ($("select#certificate-analysis").val() == "target")
          has_failed = row.target_cert_failed;
        if ($("select#certificate-analysis").val() == "page")
          has_failed = row.page_cert_failed;
      }
   
      // Count the number of times, the sensitive query succeeds in MIXED mode.
      //
      // NOTE: Since 01.02.2021, the class has a member this.completion_mode
      // (derived from the result name), which gives us the completion mode
      // explicitly. But row.is_mixed_sensitive and row.is_mixed_agnostic are
      // determined from column 14 (0-based) of the ...results.tsv file. So
      // better assert here that the two coincide.
      console.assert(!row.is_mixed_agnostic || _this.completion_mode == "mixed");
      console.assert(!row.is_mixed_sensitive || _this.completion_mode == "mixed");
      if (_this.completion_mode == "sensitive" && !row.has_failed
             || (row.is_mixed_sensitive && !sensitive_has_failed)) {
        _this.qt_all.num_queries_sensitive += 1;
        if (row.has_context)
          _this.qt_with_context.num_queries_sensitive += 1;
        else
          _this.qt_wout_context.num_queries_sensitive += 1;
      }

      // Update QT statistics.
      //
      // This is done separately for all queries, the queries with context and
      // the queries without context.
      var prefix_length = row.is_old_eval ? 0 : row.prefix_length;
      if (consider_row) {
        _this.qt_all.update(
          query_time, has_failed, is_new_token, prefix_length);
        if (row.has_context) {
          _this.qt_with_context.update(
            query_time, has_failed, is_new_token, prefix_length);
        } else {
          _this.qt_wout_context.update(
            query_time, has_failed, is_new_token, prefix_length);
        }
      }

      // Update the MRR statistics.
      //
      // NEW: this is now one histogram per prefix length, see MrrStatistics.
      // For very old eval (like 2020-08-19.hannah), skip prefix lengths > 0
      // (that's how we did it there).
      var very_old_eval_skip_row = row.is_very_old_eval && row.prefix_length > 0;
      if (consider_row && !row.is_variable_query && !very_old_eval_skip_row) {
        _this.mrr_all.update(row.position, prefix_length);
        if (row.has_context == 1) {
          _this.mrr_with_context.update(row.position, prefix_length);
        } else {
          _this.mrr_wout_context.update(row.position, prefix_length);
        }
      }

      // Update the KS statistics. This is done for each query.
      //
      // NOTE: For Fbeasy and Freebase, aliases kann be quite long, typically
      // because of disambiguation strings in paranthesis. For example:
      // The_Simpsons_(Animated_TV_Program). Hannah evaluated the effect of
      // removing the (...) from the alias when determining the max prefix
      // length. But the effect was negligible. Apparently, this is not so
      // frequent, or the problematic aliases don't have this problem.
      if (consider_row && !row.is_variable_query) {
        var max_prefix_length = row.alias.length + 1;
        // var max_prefix_length = row.alias.replace(/\s*\([^()]+\)$/, "").length + 1;
        _this.ks_all.update(
          row.prefix_length, max_prefix_length, is_new_token, row.position);
        if (row.has_context == 1) {
          _this.ks_with_context.update(
            row.prefix_length, max_prefix_length, is_new_token, row.position);
        } else {
          _this.ks_wout_context.update(
            row.prefix_length, max_prefix_length, is_new_token, row.position);
        }
      }
    });

    // Compute statistics.
    this.qt_all.compute_statistics();
    this.qt_with_context.compute_statistics();
    this.qt_wout_context.compute_statistics();
    this.mrr_all.compute_statistics();
    this.mrr_with_context.compute_statistics();
    this.mrr_wout_context.compute_statistics();
    this.ks_all.compute_statistics();
    this.ks_with_context.compute_statistics();
    this.ks_wout_context.compute_statistics();

    // console.log("#Tokens KS with context: " + this.ks_with_context.num_tokens);
    // console.log("#Tokens KS wout context: " + this.ks_wout_context.num_tokens);
  }
}
